from functools import partial
from typing import Any, Dict, Iterable
import json
from pathlib import Path
import pandas as pd
import argparse
import multiprocessing
from webdataset import TarWriter
from tqdm import tqdm


def proxy_read(data: Dict, filename_column: str):
    filename = data.pop(filename_column)
    with open(filename, 'rb') as buf:
        raw_data = buf.read()
    fpath = Path(filename)
    stem_name = str(fpath.stem).replace('.', '_')
    suffix = fpath.suffix.replace('.', '')
    ret_data = {
        suffix: raw_data,
        '__key__': f"{stem_name}",  # Just cast to str
    }
    # If we have some labels, also dump a .json file
    if len(data) > 0:
        ret_data['json'] = json.dumps(data).encode('utf-8')
    return ret_data


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'input_filelist',
        type=Path,
        help=
        "Some input filelist. We will expect a column named <filename> for the data and every other column will be dumped to json format. The filenames (basepath) need to be unique. Please first shuffle the list before processing."
    )
    parser.add_argument('outputdir', type=Path)
    parser.add_argument('-s', '--size_per_file', type=int, default=10000)
    parser.add_argument('-n', '--n_workers', type=int, default=4)
    parser.add_argument(
        '--filename_column',
        default='filename',
        type=str,
        help="The column name that identifies the files to extract")
    parser.add_argument('-d', '--delim', default='\t', type=str)
    parser.add_argument('--compress',
                        action='store_true',
                        default=False,
                        help="Using tar.gz instead of .tar")
    parser.add_argument(
        '--write_json',
        default=None,
        type=str,
        help=
        "Also writes a json to the target directory. Useful with the 'wids' library to read in random."
    )
    parser.set_defaults(stereo=False)
    args = parser.parse_args()
    df_iterator: Iterable[pd.DataFrame] = pd.read_csv(
        args.input_filelist, sep=args.delim, chunksize=args.size_per_file)

    shards_base_path = args.outputdir
    shards_base_path.mkdir(parents=True, exist_ok=True)

    suffix = '.tar' if args.compress is False else '.tar.gz'

    output_json: Dict[str, Any] = dict(wids_version=1)
    tar_file_outputs = []
    with multiprocessing.Pool(processes=args.n_workers) as pool:
        for file_num, df in enumerate(
                tqdm(df_iterator,
                     leave=True,
                     desc='Dumping to file',
                     unit='shard')):
            #Locally sample
            data = df.sample(frac=1.0).to_dict('records')
            output_file_iter = str(
                shards_base_path /
                f'{args.input_filelist.stem}_{file_num:07d}{suffix}')
            n_samples = len(data)
            tar_file_outputs.append(
                dict(url=str(output_file_iter), nsamples=n_samples))
            with TarWriter(output_file_iter,
                           encoder=False,
                           compress=args.compress) as dst:
                for return_values in tqdm(pool.imap_unordered(
                        partial(proxy_read,
                                filename_column=args.filename_column), data),
                                          unit='file',
                                          total=len(data),
                                          leave=False):
                    dst.write(return_values)
    print(f"Finished, final data can be found at {args.outputdir}")
    if args.write_json is not None:
        import json
        output_json['shardlist'] = tar_file_outputs
        with open(args.write_json, 'w') as f:
            json.dump(output_json, f)
        print(f"Dumped Json for wids usage at {args.write_json}")



from __future__ import annotations
from pathlib import Path
import uuid
from typing import Dict, List, Optional, Tuple, Any, Type

import yaml
import datetime
import torch
from dataclasses import dataclass, field, asdict


@dataclass
class MAEConfig:
    train_data: List[str]
    cv_data: List[str]
    config_file: str = ''  # Will be overwritten during parsing
    logfile: str = 'train.log'
    outputpath: str = 'experiments'
    #Train args
    mask_ratio: float = 0.75
    use_scheduler: bool = True
    warmup_iters: Optional[int] = None
    warmup_epochs: Optional[int] = None
    model: str = 'dasheng_base'
    model_args: Dict[str, Any] = field(default_factory=lambda: dict())
    decay_frac: float = 0.01  # Decay fraction  of learning rate

    optimizer: str = 'AdamW8bit'
    optimizer_args: Dict[str,
                         Any] = field(default_factory=lambda: dict(lr=0.0003, weight_decay=0.01))
    epochs: int = 100
    epoch_length: int = 15000
    # Dataloader args
    batch_size: int = 32
    n_saved: int = 4  # Num models saved
    num_workers: int = 4
    resampled: bool = True
    crop_shuffle: int = 512
    chunk_length: float = 10.0  # Sample length during training/testing
    sample_rate: int = 16000  # Sampling rate of audio
    valid_every: int = 1  # When to run validation
    score_function: Tuple[str, float] = ('Loss', -1.0) # Save best loss on CV
    average_final_model: bool = True 
    outputdir: Path = field(init=False)

    def __post_init__(self):
        self.outputdir = Path(self.outputpath) / Path(
            self.config_file
        ).stem / self.model / f"{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')}_{uuid.uuid1().hex}"

    def to_dict(self):
        return asdict(self)

    def state_dict(self):
        return self.to_dict()

    @classmethod
    def load_state_dict(cls, state):
        return cls(**state)

    @classmethod
    def from_config_file(cls: Type[MAEConfig], config_file: str,
                         **kwargs) -> MAEConfig:
        """parse_config_or_kwargs

        :param config_file: Config file that has parameters, yaml format
        :param **kwargs: Other alternative parameters or overwrites for config
        """
        with open(config_file) as con_read:
            yaml_config = yaml.load(con_read, Loader=yaml.FullLoader)
        # values from config file are all possible params
        return cls(**dict(yaml_config, config_file=config_file, **kwargs))


def average_models(models: List[str]):
    model_res_state_dict = {}
    state_dict = {}
    has_new_structure = False
    for m in models:
        cur_state = torch.load(m, map_location='cpu')
        if 'model' in cur_state:
            has_new_structure = True
            model_params = cur_state.pop('model')
            # Append non "model" items, encoder, optimizer etc ...
            for k in cur_state:
                state_dict[k] = cur_state[k]
            # Accumulate statistics
            for k in model_params:
                if k in model_res_state_dict:
                    model_res_state_dict[k] += model_params[k]
                else:
                    model_res_state_dict[k] = model_params[k]
        else:
            for k in cur_state:
                if k in model_res_state_dict:
                    model_res_state_dict[k] += cur_state[k]
                else:
                    model_res_state_dict[k] = cur_state[k]

    # Average
    for k in model_res_state_dict:
        # If there are any parameters
        if model_res_state_dict[k].ndim > 0:
            model_res_state_dict[k] /= float(len(models))
    if has_new_structure:
        state_dict['model'] = model_res_state_dict
    else:
        state_dict = model_res_state_dict
    return state_dict



import torch
from einops import rearrange
from ..train.models import AudioTransformerMAE_Encoder

PRETRAINED_CHECKPOINTS = {
    'dasheng_base':
    'https://zenodo.org/records/11511780/files/dasheng_base.pt?download=1',
    'dasheng_06B':
    'https://zenodo.org/records/11511780/files/dasheng_06b.pt?download=1',
    'dasheng_12B':
    'https://zenodo.org/records/11511780/files/dasheng_12b.pt?download=1',
}


#Using the pretrained encoders, remove all masking
class Dasheng(AudioTransformerMAE_Encoder):

    # need the *args, **kwargs otherwise we get a linter warning
    def forward_features(self, x: torch.Tensor, *args,
                         **kwargs) -> torch.Tensor:
        x = self.patch_embed(x)
        *_, t = x.shape
        x = x + self.time_pos_embed[:, :, :, :t]
        x = x + self.freq_pos_embed[:, :, :, :]
        x = rearrange(x, 'b c f t -> b (f t) c')
        if self.pooling == 'token':
            cls_token = self.cls_token.expand(x.shape[0], -1, -1)
            cls_token = cls_token + self.token_pos_embed[:, :]
            x = torch.cat((cls_token, x), dim=1)
        x = self.pos_drop(x)
        x = self.blocks(x)
        x = self.norm(x)
        return x

    def forward_spectrogram(self, x: torch.Tensor) -> torch.Tensor:
        if x.shape[-1] > self.target_length:
            splits = x.split(self.target_length, -1)

            padding_start_in_chunks = 0
            if len(splits) > 1 and (splits[-1].shape[-1] < self.target_length):
                pad = torch.zeros(*x.shape[:-1],
                                  self.target_length,
                                  device=x.device)
                pad[..., :splits[-1].shape[-1]] = splits[-1]
                padding_start_in_chunks = splits[-1].shape[-1] // self.patch_stride[-1]
                splits = torch.stack((*splits[:-1], pad), dim=0)
            n_splits = len(splits)
            # Splits into the batch_size, speeds up computation
            x = rearrange(splits, 'spl b c f t-> (spl b) c f t')
            x = self.forward_features(x)
            x = rearrange(x, '(spl b) ... d -> spl b (...) d', spl=n_splits)
            # Trim last chunk
            if padding_start_in_chunks > 0:
                not_padded = rearrange(x[:-1], 'spl b ... d -> b (spl ...) d')
                padded = x[-1][:, :padding_start_in_chunks, ...]
                x = torch.cat((not_padded, padded), dim=1)
            else:
                x = rearrange(x, 'spl b ... d -> b (spl ...) d', spl=n_splits)
            return x
        else:
            return self.forward_features(x)



    def forward(self, x, *args, **kwargs) -> torch.Tensor:
        x = self.forward_to_spec(x)
        return self.forward_spectrogram(x)

    @classmethod
    def from_pretrained(
            cls, pretrained_url: str,
            **additional_model_kwargs) -> AudioTransformerMAE_Encoder:
        """
        Class method to create a new Vocos model instance from a pre-trained model stored in the Hugging Face model hub.
        """
        if 'http' in pretrained_url:
            dump = torch.hub.load_state_dict_from_url(pretrained_url,
                                                      map_location='cpu')
        else:
            dump = torch.load(pretrained_url, map_location='cpu')
        model_parmeters, model_config = dump['model'], dump['config']
        instance = cls(**{**model_config, **additional_model_kwargs})
        instance.load_state_dict(model_parmeters, strict=True)
        return instance


def dasheng_base(**model_kwargs):
    return Dasheng.from_pretrained(PRETRAINED_CHECKPOINTS['dasheng_base'],
                                   **model_kwargs)


def dasheng_06B(**model_kwargs):
    return Dasheng.from_pretrained(PRETRAINED_CHECKPOINTS['dasheng_06B'],
                                   **model_kwargs)


def dasheng_12B(**model_kwargs):
    return Dasheng.from_pretrained(PRETRAINED_CHECKPOINTS['dasheng_12B'],
                                   **model_kwargs)


if __name__ == "__main__":
    mdl = dasheng_base()
    print(mdl(torch.randn(1, 168499)).shape)



from loguru import logger
from fire import Fire
import numpy as np
from audiowebdataset import create_dataloader

import models
import utils
import torch
import sys
import ignite
from ignite.contrib.handlers import ProgressBar
from ignite.engine import (Engine, Events)
from ignite.handlers import (Checkpoint, DiskSaver, global_step_from_engine,
                             create_lr_scheduler_with_warmup)
from accelerate import Accelerator

logger.configure(handlers=[{
    "sink": sys.stderr,
    "format": "[<green>{time:YYYY-MM-DD HH:mm:ss}</green>] {message}",
    'level': 'DEBUG',
}])


def transfer_to_device(batch, device):
    return (x.to(device, non_blocking=True)
            if isinstance(x, torch.Tensor) else x for x in batch)


def create_engine(engine_function,
                  local_rank: int = 0,
                  output_transform=lambda x: x):
    engine = Engine(engine_function)
    if local_rank == 0:
        ProgressBar().attach(engine, output_transform=output_transform)
    return engine


class Runner(object):

    def __init__(self, seed: int = 42, nthreads: int = 1):
        super().__init__()
        self.seed = seed
        torch.manual_seed(seed)
        np.random.seed(seed)
        torch.set_num_threads(nthreads)
        logger.info(f"Using seed {seed}")

    def __create_dir(self, config: utils.MAEConfig):
        config.outputdir.mkdir(exist_ok=True, parents=True)
        logger.add(
            config.outputdir / config.logfile,
            enqueue=True,
            level='INFO',
            format=
            "[<red>{level}</red> <green>{time:YYYY-MM-DD HH:mm:ss}</green>] {message}"
        )

    def log_basic_info(self, config_parameters: utils.MAEConfig, device):
        logger.info(f"Running on device {device}")
        logger.info(f"Storing output in {config_parameters.outputdir}")
        logger.info(f"- PyTorch version: {torch.__version__}")
        logger.info(f"- Ignite version: {ignite.__version__}")
        if torch.cuda.is_available():
            logger.info(f"- GPU Device: {torch.cuda.current_device()}")
            logger.info(f"- CUDA version: {torch.version.cuda}")
        for k, v in config_parameters.to_dict().items():
            logger.info(f"{k} : {v}")

    def train(self, config, **overwrite_kwargs):
        config_parameters = utils.MAEConfig.from_config_file(
            config, **overwrite_kwargs)
        accelerator = Accelerator()

        def log(message: str):
            if accelerator.is_main_process:
                logger.info(message)

        if accelerator.is_main_process:
            self.__create_dir(config_parameters)
            self.log_basic_info(config_parameters, device=accelerator.device)
        train_dataloader = create_dataloader(
            config_parameters.train_data,
            crop_size=int(config_parameters.chunk_length *
                          config_parameters.sample_rate),
            batch_size=config_parameters.batch_size,
            crop_shuffle=config_parameters.crop_shuffle,
            resampled=True)

        test_dataloader = create_dataloader(
            config_parameters.cv_data,
            crop_size=int(config_parameters.chunk_length *
                          config_parameters.sample_rate),
            batch_size=config_parameters.batch_size)

        model = getattr(
            models,
            config_parameters.model)(**config_parameters.model_args).train()
        log(model)

        if '8bit' in config_parameters.optimizer:
            import bitsandbytes as bnb
            optimizer = getattr(bnb.optim, config_parameters.optimizer)(
                model.parameters(),
                **config_parameters.optimizer_args)  # add bnb optimizer
        else:
            optimizer = getattr(torch.optim, config_parameters.optimizer)(
                model.parameters(), **config_parameters.optimizer_args)

        def _inference(engine, batch):
            model.eval()
            with torch.no_grad():
                with accelerator.autocast():
                    x, *_ = transfer_to_device(batch, accelerator.device)
                    loss = model(x,
                                 mask_ratio=config_parameters.mask_ratio,
                                 return_loss=True)
                    return loss

        def train_batch(engine, batch):
            model.train()
            with torch.enable_grad():
                x, *_ = transfer_to_device(batch, accelerator.device)
                optimizer.zero_grad()
                with accelerator.autocast():
                    loss = model(x,
                                 mask_ratio=config_parameters.mask_ratio,
                                 return_loss=True)
                    accelerator.backward(loss)
                    optimizer.step()
                    return {
                        'loss': loss.item(),
                        'lr': optimizer.param_groups[0]['lr']
                    }

        def run_validation(engine, title=None):
            if accelerator.is_main_process:
                results = engine.state.metrics
                output_str_list = [
                    f"{title:<10} Results - Epoch : {train_engine.state.epoch:<4}"
                ] + [
                    f"{metric} {results[metric]:<5.4f}" for metric in results
                ] + [f"LR: {optimizer.param_groups[0]['lr']:.2e}"]
                log(" ".join(output_str_list))

        train_engine = create_engine(train_batch)
        inference_engine = create_engine(_inference, output_transform=None)
        ignite.metrics.Average().attach(inference_engine, 'Loss')

        score_function = Checkpoint.get_default_score_fn(
            *config_parameters.score_function)
        checkpoint_saver = Checkpoint(
            {
                'model': model.encoder,
                'config': config_parameters,
            },
            DiskSaver(config_parameters.outputdir),
            n_saved=config_parameters.n_saved,
            global_step_transform=global_step_from_engine(train_engine),
            filename_prefix='best',
            score_function=score_function)
        last_checkpoint_saver = Checkpoint(
            {
                'model': model.encoder,
                'config': config_parameters
            },
            DiskSaver(config_parameters.outputdir),
            n_saved=1,
            global_step_transform=global_step_from_engine(train_engine))

        train_length = config_parameters.epoch_length * config_parameters.epochs
        decay_steps = train_length

        if config_parameters.use_scheduler:
            scheduler = ignite.handlers.param_scheduler.CosineAnnealingScheduler(
                optimizer, 'lr', optimizer.param_groups[0]['lr'],
                optimizer.param_groups[0]['lr'] * config_parameters.decay_frac,
                decay_steps)
            warmup_time_in_iters = None
            if config_parameters.warmup_iters is not None:
                warmup_time_in_iters = config_parameters.warmup_iters
            elif config_parameters.warmup_epochs is not None:
                warmup_time_in_iters = config_parameters.epoch_length * config_parameters.warmup_epochs
            if warmup_time_in_iters is not None:
                log(f"Using warmup with {warmup_time_in_iters} iters")
                scheduler = create_lr_scheduler_with_warmup(
                    scheduler,
                    warmup_start_value=0.0,
                    warmup_duration=warmup_time_in_iters)

            train_engine.add_event_handler(Events.ITERATION_STARTED, scheduler)
        inference_engine.add_event_handler(Events.COMPLETED, checkpoint_saver)
        inference_engine.add_event_handler(Events.COMPLETED,
                                           last_checkpoint_saver)

        @train_engine.on(
            Events.EPOCH_COMPLETED(every=config_parameters.valid_every))
        def valid_eval(train_engine):
            with inference_engine.add_event_handler(Events.COMPLETED,
                                                    run_validation,
                                                    "Validation"):
                inference_engine.run(test_dataloader)

        model, optimizer, train_dataloader, test_dataloader = accelerator.prepare(
            model, optimizer, train_dataloader, test_dataloader)

        train_engine.run(
            train_dataloader,
            max_epochs=config_parameters.epochs,
            epoch_length=config_parameters.epoch_length,
        )
        output_model = config_parameters.outputdir / checkpoint_saver.last_checkpoint
        if config_parameters.average_final_model:
            log("Averaging best models ...")
            output_model = config_parameters.outputdir / 'averaged.pt'

            averaged_state_dict = utils.average_models([
                config_parameters.outputdir / f.filename
                for f in checkpoint_saver._saved
            ])
            torch.save(averaged_state_dict, output_model)


if __name__ == "__main__":
    Fire(Runner().train)



from typing import List, Optional
import random
from functools import partial
import webdataset as wds
import braceexpand # Dependency in wds
import torch
import numpy as np

def crop_or_pad(wav: torch.Tensor, crop_size: int, pad_last: bool = False):
    n_samples, *_ = wav.shape
    available_crops = n_samples // crop_size
    for i in range(available_crops):
        crop = wav[i * crop_size:(i + 1) * crop_size, ...]
        yield crop

    if (available_crops == 0) or (pad_last):
        last_crop = wav[available_crops * crop_size:, ...]
        padded = torch.zeros((crop_size, *last_crop.shape[1:]))
        padded[:last_crop.shape[0]] = last_crop
        yield padded


def convert_decibels_to_amplitude_ratio(decibels):
    return 10**(decibels / 20)


def _audio_gain(data_stream, min_gain_db: float = -6, max_gain_db=10):
    for sample in data_stream:
        audio, *extra = sample
        scale_factor = convert_decibels_to_amplitude_ratio(
            random.uniform(min_gain_db, max_gain_db))
        yield (audio * scale_factor, *extra)

def _seq_crop(data, crop_size: int, mono: bool = True, pad_last: bool = False, drop_crops: bool = False,handler=None):
    """WebDataset crop filter, yields sequential crops"""
    for sample in data:
        audio, *extra = sample
        if isinstance(audio, tuple):
            audio = audio[0]
        if mono and audio.ndim == 2:
            audio = audio.mean(0)
        if drop_crops and audio.shape[-1] < int(crop_size * 0.8):
            continue
        crops = crop_or_pad(audio.float(),
                            crop_size=crop_size,
                            pad_last=pad_last)
        for crop in crops:
            yield (crop, *extra)


class Audiowebdataset_Fluid(wds.DataPipeline):

    def __init__(self,
                 urls,
                 shuffle: Optional[int] = None,
                 crop_size: int = 16000,
                 resample: bool = False,
                 crop_shuffle: Optional[int] = None,
                 batch_size: Optional[int] = None,
                 add_gain: bool = False,
                 drop_crops: bool = False,

                 ):
        pipeline: List = [
            wds.SimpleShardList(urls)
            if resample is False else wds.ResampledShards(urls)
        ]
        if shuffle is not None:
            # Tar wise shuffle
            pipeline.extend([
                wds.detshuffle(
                    bufsize=shuffle,
                    initial=shuffle // 4,
                ),
                wds.split_by_node,
                wds.split_by_worker,
                # at this point, we have an iterator over the shards assigned to each worker at each node
                wds.tarfile_to_samples(handler=wds.warn_and_continue),
                wds.shuffle(
                    bufsize=shuffle,
                    initial=shuffle // 4,
                ),
            ])
        else:
            pipeline.extend([wds.split_by_worker, wds.tarfile_to_samples()])
        pipeline.extend([
            wds.decode(wds.torch_audio, handler=wds.warn_and_continue),
            wds.to_tuple("mp3;wav;flac", "__key__"),
            partial(_seq_crop, crop_size=crop_size, drop_crops = drop_crops)
        ])
        if add_gain:
            pipeline.extend([_audio_gain])
        if crop_shuffle is not None:
            pipeline.append(wds.shuffle(crop_shuffle))
        if batch_size is not None:
            pipeline.append(wds.batched(batch_size))
        super().__init__(pipeline)


#Can also replace with wds.Randomix
class SampleDatasets(wds.DataPipeline, wds.compat.FluidInterface):

    def __init__(self, datasets, probability: Optional[List[float]] = None):
        super().__init__()
        self.datasets = datasets
        if probability is None:
            probability = [1.0] * len(self.datasets)
        self.prob = probability

    def __iter__(self):
        sources = [iter(ds) for ds in self.datasets]
        while True:
            for source in random.choices(sources, weights=self.prob):
                try:
                    yield next(source)
                except StopIteration:
                    return


def create_dataloader(data_urls: List[str],
                      crop_size: int,
                      batch_size: int = 32,
                      crop_shuffle: Optional[int] = None,
                      resampled: bool = False,
                      num_workers: int = 4,
                      *args,
                      **kwargs):
    train_lists: List[str] = []
    for train_data_url in data_urls:
        train_lists.extend(braceexpand.braceexpand(train_data_url))
    ds = Audiowebdataset_Fluid(
        train_lists,
        crop_size=crop_size,
        resample=resampled,
        batch_size=batch_size,
        crop_shuffle=crop_shuffle,
        shuffle=crop_shuffle,
    )
    dataloader = wds.WebLoader(ds, batch_size=None, num_workers=num_workers)
    if crop_shuffle is not None:
        dataloader = dataloader.unbatched().shuffle(crop_shuffle).batched(
            batch_size)
    return dataloader



import torch
import torch.nn as nn
from einops import rearrange
import torch
from torch.cuda.amp import autocast
from functools import partial
from typing import Optional, Tuple, Union
import torchaudio.transforms as audio_transforms
from einops import rearrange
from einops.layers.torch import Rearrange
from itertools import repeat
import collections


def _ntuple(n):

    def parse(x) -> Tuple:
        if isinstance(x, collections.abc.Iterable) and not isinstance(x, str):
            return tuple(x)
        return tuple(repeat(x, n))

    return parse


to_2tuple = _ntuple(2)


class MAELoss(torch.nn.Module):

    def __init__(self, norm_pix_loss: bool = True):
        super().__init__()
        self.norm_pix_loss = norm_pix_loss

    @autocast(enabled=False)
    def forward(self, pred: torch.Tensor, target: torch.Tensor,
                mask: torch.Tensor) -> torch.Tensor:
        if self.norm_pix_loss is True:
            mean = target.mean(dim=-1, keepdim=True)
            var = target.var(dim=-1, keepdim=True)
            target = (target - mean) / (var + 1.e-6)**.5
        elif self.norm_pix_loss == 'global':
            mean = target.mean()
            var = target.var()
            target = (target - mean) / (var + 1.e-6)**.5
        loss = (pred - target)**2
        loss = loss.mean(dim=-1)  # [N, L], mean loss per patch
        loss = (loss * mask).sum() / mask.sum()  # mean loss on removed patches
        return loss


class AudioPatchEmbed(nn.Module):

    def __init__(self,
                 input_size: Union[int, Tuple[int, int]] = (64, 100),
                 patch_size: Tuple[int, int] = (64, 4),
                 patch_stride: Tuple[int, int] = (64, 4),
                 in_chans=1,
                 embed_dim=768,
                 norm_layer=None,
                 flatten=False):
        super().__init__()
        patch_size = to_2tuple(patch_size)
        patch_stride = to_2tuple(patch_stride)
        self.input_size: Tuple[int, int] = to_2tuple(input_size)
        self.patch_size: Tuple[int, int] = to_2tuple(patch_size)
        self.patch_stride: Tuple[int, int] = to_2tuple(patch_stride)
        self.grid_size = (self.input_size[0] // self.patch_stride[0],
                          self.input_size[1] // self.patch_stride[1])
        self.num_patches = self.grid_size[0] * self.grid_size[1]
        self.flatten = flatten

        self.proj = nn.Conv2d(in_chans,
                              embed_dim,
                              kernel_size=patch_size,
                              stride=patch_stride)
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x):
        x = self.proj(x)
        if self.flatten:
            x = rearrange(x, 'b c f t -> b (f t) c')
        x = self.norm(x)
        return x


class LayerScale(nn.Module):

    def __init__(self, dim: int, init_values=1e-5, inplace=False):
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x):
        return x.mul_(self.gamma) if self.inplace else x * self.gamma


class Attention(nn.Module):

    def __init__(self,
                 dim,
                 num_heads=8,
                 qkv_bias=False,
                 attn_drop=0.,
                 proj_drop=0.):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads,
                                  C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(
            0)  # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Mlp(nn.Module):

    def __init__(self,
                 in_features,
                 hidden_features=None,
                 out_features=None,
                 act_layer=nn.GELU,
                 drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Block(nn.Module):

    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.,
        qkv_bias=False,
        drop=0.,
        attn_drop=0.,
        init_values=None,
        act_layer: torch.Callable = nn.GELU,
        norm_layer: torch.Callable = nn.LayerNorm,
        attention_type='Attention',
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        attn_type = globals()[attention_type]
        self.attn = attn_type(dim,
                              num_heads=num_heads,
                              qkv_bias=qkv_bias,
                              attn_drop=attn_drop,
                              proj_drop=drop)
        self.ls1 = LayerScale(
            dim, init_values=init_values) if init_values else nn.Identity()

        self.norm2 = norm_layer(dim)
        self.mlp = Mlp(in_features=dim,
                       hidden_features=int(dim * mlp_ratio),
                       act_layer=act_layer,
                       drop=drop)
        self.ls2 = LayerScale(
            dim, init_values=init_values) if init_values else nn.Identity()

    def forward(self, x):
        x = x + self.ls1(self.attn(self.norm1(x)))
        x = x + self.ls2(self.mlp(self.norm2(x)))
        return x


class AudioTransformerMAE_Encoder(nn.Module):

    def __init__(self,
                 patch_size: Tuple[int, int] = (64, 4),
                 patch_stride: Tuple[int, int] = (64, 4),
                 embed_dim: int = 768,
                 depth: int = 12,
                 num_heads=8,
                 mlp_ratio=4.,
                 qkv_bias=True,
                 drop_rate=0.,
                 attn_drop_rate=0.,
                 norm_layer=None,
                 act_layer=None,
                 init_values=None,
                 target_length=1008,
                 pooling='mean',
                 time_patch_out: Optional[float] = None,
                 freq_patch_out: Optional[float] = None,
                 block_type='Block',
                 attention_type='Attention',
                 eval_avg='cat',
                 n_fft: int = 512,
                 n_mels: int = 64,
                 hop_size: int = 160,
                 win_size: int = 512,
                 f_min: int = 0,
                 f_max: int = 8000,
                 center: bool = True,
                 **kwargs):
        super().__init__()
        self.pooling = pooling
        self.embed_dim = embed_dim
        self.patch_stride = patch_stride
        self.patch_size = patch_size
        self.n_mels = n_mels
        self.eval_avg = eval_avg
        self.time_patch_out = time_patch_out
        self.freq_patch_out = freq_patch_out

        self.front_end = nn.Sequential(
            audio_transforms.MelSpectrogram(f_min=f_min,
                                            sample_rate=16000,
                                            win_length=win_size,
                                            center=center,
                                            n_fft=n_fft,
                                            f_max=f_max,
                                            hop_length=hop_size,
                                            n_mels=self.n_mels),
            audio_transforms.AmplitudeToDB(top_db=kwargs.get('top_db', 120)))

        self.init_bn = nn.Sequential(
            Rearrange('b c f t -> b f c t'),
            nn.BatchNorm2d(self.n_mels, momentum=0.01),
            Rearrange('b f c t -> b c f t'))

        self.target_length = target_length
        self.patch_embed = AudioPatchEmbed(input_size=(self.n_mels,
                                                       target_length),
                                           embed_dim=self.embed_dim,
                                           patch_size=self.patch_size,
                                           flatten=False,
                                           patch_stride=self.patch_stride)
        self.num_patches = self.patch_embed.num_patches

        if pooling == 'token':
            self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
            self.token_pos_embed = nn.Parameter(
                torch.randn(1, embed_dim) * .02)

        self.time_pos_embed = nn.Parameter(
            torch.randn(1, embed_dim, 1, self.patch_embed.grid_size[1]) * .02)
        self.freq_pos_embed = nn.Parameter(
            torch.randn(1, embed_dim, self.patch_embed.grid_size[0], 1) * .02)

        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        act_layer = act_layer or nn.GELU
        self.pos_drop = nn.Dropout(p=drop_rate)
        block_function = globals()[block_type]
        self.blocks = nn.Sequential(*[
            block_function(
                dim=embed_dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                init_values=init_values,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                norm_layer=norm_layer,
                act_layer=act_layer,
                attention_type=attention_type,
            ) for _ in range(depth)
        ])
        self.norm = norm_layer(embed_dim)
        self.apply(self.init_weights)
        if hasattr(self, 'cls_token') and self.cls_token is not None:
            nn.init.normal_(self.cls_token, std=1e-6)
        group_masking = kwargs.get('group_masking', False)
        if isinstance(group_masking, bool):
            if group_masking is True:
                self.masking_func = self.random_masking_group
            else:
                self.masking_func = self.random_masking
        elif isinstance(group_masking, int):
            self.masking_func = partial(self.random_masking_group,
                                        group_factor=group_masking)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {
            'time_pos_embed', 'cls_token', 'freq_pos_embed', 'token_pos_embed'
        }

    def init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.xavier_uniform_(module.weight)
            if module.bias is not None:
                nn.init.zeros_(module.bias)
        elif isinstance(module, nn.LayerNorm):
            nn.init.constant_(module.bias, 0)
            nn.init.constant_(module.weight, 1.0)

    def random_masking_group(self, x, mask_ratio, group_factor: int = 2):
        """                                                                                                                                                                                       
        Perform per-sample random masking by per-sample shuffling.                                                                                                                                
        Per-sample shuffling is done by argsort random noise.                                                                                                                                     
        x: [N, L, D], sequence                                                                                                                                                                    
        """
        N, L, D = x.shape  # batch, length, dim
        len_keep = int(L * (1 - mask_ratio))

        noise = torch.rand(N, L // group_factor,
                           device=x.device)  # noise in [0, 1]
        # indices = torch.arange(L).view(1, 5, 4).repeat(N, 1, 1)
        indices = torch.arange(L, device=x.device).view(-1, group_factor)

        # sort noise for each sample
        ids_shuffle = torch.argsort(
            noise, dim=1)  # ascend: small is keep, large is remove
        ids_shuffle = indices[ids_shuffle].flatten(-2)
        ids_restore = torch.argsort(ids_shuffle, dim=1)

        # keep the first subset
        ids_keep = ids_shuffle[:, :len_keep]
        x_masked = torch.gather(x,
                                dim=1,
                                index=ids_keep.unsqueeze(-1).repeat(1, 1, D))

        # generate the binary mask: 0 is keep, 1 is remove
        mask = torch.ones([N, L], device=x.device)
        mask[:, :len_keep] = 0
        # unshuffle to get the binary mask
        mask = torch.gather(mask, dim=1, index=ids_restore)

        return x_masked, mask, ids_restore

    def random_masking(self, x, mask_ratio):
        """
        Perform per-sample random masking by per-sample shuffling.
        Per-sample shuffling is done by argsort random noise.
        x: [N, L, D], sequence
        """
        N, L, D = x.shape  # batch, length, dim
        len_keep = int(L * (1 - mask_ratio))

        noise = torch.rand(N, L, device=x.device)  # noise in [0, 1]

        # sort noise for each sample
        ids_shuffle = torch.argsort(
            noise, dim=1)  # ascend: small is keep, large is remove
        ids_restore = torch.argsort(ids_shuffle, dim=1)

        # keep the first subset
        ids_keep = ids_shuffle[:, :len_keep]
        x_masked = torch.gather(x,
                                dim=1,
                                index=ids_keep.unsqueeze(-1).repeat(1, 1, D))

        # generate the binary mask: 0 is keep, 1 is remove
        mask = torch.ones([N, L], device=x.device)
        mask[:, :len_keep] = 0
        # unshuffle to get the binary mask
        mask = torch.gather(mask, dim=1, index=ids_restore)

        return x_masked, mask, ids_restore

    def forward_features(self, x, mask_ratio):
        x = self.patch_embed(x)
        b, c, f, t = x.shape
        x = x + self.time_pos_embed[:, :, :, :t]
        x = x + self.freq_pos_embed[:, :, :, :]  # Just for sin pos embed
        x = rearrange(x, 'b c f t -> b (f t) c')
        # x, mask, ids_restore = self.random_masking(x, mask_ratio)
        x, mask, ids_restore = self.masking_func(x, mask_ratio)
        if self.pooling == 'token':
            cls_token = self.cls_token.expand(x.shape[0], -1, -1)
            cls_token = cls_token + self.token_pos_embed[:, :]
            x = torch.cat((cls_token, x), dim=1)
        x = self.pos_drop(x)
        x = self.blocks(x)
        x = self.norm(x)
        return x, mask, ids_restore

    def load_state_dict(self, state_dict, strict=True, **kwargs):
        if 'time_pos_embed' in state_dict and self.time_pos_embed.shape != state_dict[
                'time_pos_embed'].shape:
            print(
                "Positional Embedding shape not the same with model, resizing!"
            )
            self.change_pos_embedding(state_dict)
        super().load_state_dict(state_dict, strict=strict, **kwargs)

    def change_pos_embedding(self, state_dict):
        target_time_pos_embed_length = self.time_pos_embed.shape[-1]
        target_freq_pos_embed_length = self.freq_pos_embed.shape[-2]

        pretrained_time_pos_embed = state_dict['time_pos_embed']
        pretrained_freq_pos_embed = state_dict['freq_pos_embed']

        if target_freq_pos_embed_length <= pretrained_time_pos_embed.shape[-1]:
            state_dict['time_pos_embed'] = pretrained_time_pos_embed[
                ..., :target_time_pos_embed_length]
        else:
            state_dict['time_pos_embed'] = torch.nn.functional.interpolate(
                pretrained_time_pos_embed,
                size=(1, target_time_pos_embed_length),
                align_corners=False,
                mode='bilinear')
        if target_freq_pos_embed_length <= pretrained_freq_pos_embed.shape[-2]:
            state_dict[
                'freq_pos_embed'] = pretrained_freq_pos_embed[:, :, :
                                                              target_freq_pos_embed_length, :]
        else:
            state_dict['freq_pos_embed'] = torch.nn.functional.interpolate(
                pretrained_freq_pos_embed,
                size=(target_freq_pos_embed_length, 1),
                align_corners=False,
                mode='bilinear')

    def forward_to_spec(self, x):
        # Do not use fp16 for feature extraction, that is likely to get nan
        with autocast(enabled=False):
            X = self.front_end(x)
            X = rearrange(X, 'b f t -> b 1 f t')
            X = self.init_bn(X)
        return X

    def forward(self, x, mask_ratio: float = 0.75):
        x = self.forward_to_spec(x)
        x, mask, restore_idxs = self.forward_features(x, mask_ratio=mask_ratio)
        return x, mask, restore_idxs


class AudioTransformerMAE_Decoder(nn.Module):

    def __init__(self,
                 input_dim: int,
                 outputdim: int,
                 patch_size: int = 16,
                 patch_stride: int = 16,
                 embed_dim: int = 768,
                 num_patches: int = 100,
                 depth: int = 12,
                 num_heads: int = 12,
                 mlp_ratio: float = 4.,
                 qkv_bias: bool = True,
                 drop_rate: float = 0.,
                 attn_drop_rate: float = 0.,
                 norm_layer: Optional[torch.nn.Module] = None,
                 act_layer: Optional[torch.nn.Module] = None,
                 cls_token: bool = False,
                 attention_type='Attention',
                 init_values=None,
                 **kwargs):
        super().__init__()
        self.embed_dim = embed_dim
        self.patch_stride = patch_stride
        self.patch_size = patch_size
        self.input_dim = input_dim

        self.input_proj = nn.Linear(input_dim, embed_dim)

        self.mask_token = nn.Parameter(torch.randn(1, 1, embed_dim) * .02)
        _norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        _act_layer = act_layer or nn.GELU
        self.use_cls = cls_token
        num_patches_total = num_patches + 1 if not cls_token else num_patches
        self.pos_embed = nn.Parameter(
            torch.zeros(1, num_patches_total, embed_dim))
        self.pos_drop = nn.Dropout(p=drop_rate)
        self.blocks = nn.Sequential(*[
            Block(
                dim=embed_dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                init_values=init_values,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                norm_layer=_norm_layer,
                act_layer=_act_layer,
                attention_type=attention_type,
            ) for i in range(depth)
        ])
        self.norm = _norm_layer(embed_dim)
        self.outputlayer = nn.Linear(self.embed_dim, outputdim)
        self.apply(self.init_weights)
        torch.nn.init.normal_(self.mask_token, std=.02)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {
            'time_pos_embed', 'cls_token', 'freq_pos_embed', 'token_pos_embed'
        }

    def init_weights(self, module):
        if isinstance(module, nn.Linear):
            nn.init.trunc_normal_(module.weight, std=.02)
            if module.bias is not None:
                nn.init.zeros_(module.bias)
        elif isinstance(module, nn.LayerNorm):
            nn.init.constant_(module.bias, 0)
            nn.init.constant_(module.weight, 1.0)

    def forward_features(self, x, ids_restore):
        x = self.input_proj(x)
        mask_tokens = self.mask_token.repeat(
            x.shape[0], ids_restore.shape[1] + 1 - x.shape[1], 1)
        if self.use_cls:
            x_ = torch.cat([x[:, 1:, :], mask_tokens], dim=1)  # no cls token
        else:
            x_ = torch.cat([x[:, :, :], mask_tokens], dim=1)
        x_ = torch.gather(x_,
                          dim=1,
                          index=ids_restore.unsqueeze(-1).repeat(
                              1, 1, x.shape[2]))  # unshuffle
        if self.use_cls:
            x = torch.cat([x[:, :1, :], x_], dim=1)  # append cls token
        else:
            x = x_
        t = x.shape[1]

        x = x + self.pos_embed[:, :t, :]
        x = self.pos_drop(x)
        x = self.blocks(x)
        x = self.norm(x)
        return x

    def forward(self, x, restore_idxs):
        x = self.forward_features(x, restore_idxs)
        x = self.outputlayer(x)
        return x


class AudioTransformerMAE(nn.Module):

    def __init__(self,
                 encoder: AudioTransformerMAE_Encoder,
                 decoder: AudioTransformerMAE_Decoder,
                 loss_fn: Optional[torch.nn.Module] = None):
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.unfold = nn.Unfold(
            kernel_size=self.encoder.patch_embed.patch_size,
            stride=self.encoder.patch_embed.patch_size)
        self.loss_fn = MAELoss() if loss_fn is None else loss_fn

    def forward(self,
                x: torch.Tensor,
                mask_ratio: float = 0.75,
                return_loss: bool = False):
        latent, mask, restore_ids = self.encoder(x, mask_ratio=mask_ratio)
        pred = self.decoder(latent, restore_ids)
        with autocast(enabled=False):
            targets = self.encoder.front_end(x)
        targets = self.patchify(targets)
        if return_loss:
            return self.loss_fn(pred, targets, mask)
        return pred, targets, mask

    def patchify(self, x):
        return self.unfold(x.unsqueeze(1)).transpose(-2, -1)


def dasheng_base(**kwargs):
    encoder_kwargs = dict(embed_dim=768,
                          depth=12,
                          num_heads=12,
                          target_length=1008,
                          patch_size=[64, 4],
                          patch_stride=[64, 4])
    encoder_kwargs.update(
        (k, kwargs[k]) for k in set(kwargs).intersection(encoder_kwargs))
    encoder_kwargs = {**encoder_kwargs, **kwargs}
    encoder = AudioTransformerMAE_Encoder(**encoder_kwargs)

    decoder_kwargs = dict(embed_dim=512,
                          depth=8,
                          num_heads=16,
                          input_dim=encoder_kwargs['embed_dim'],
                          outputdim=encoder.patch_embed.patch_size[0] *
                          encoder.patch_embed.patch_size[1],
                          num_patches=encoder.patch_embed.num_patches)
    decoder = AudioTransformerMAE_Decoder(**decoder_kwargs)
    return AudioTransformerMAE(encoder, decoder)


def dasheng_06B(**kwargs):
    encoder_kwargs = dict(
        patch_size=[64, 4],
        patch_stride=[64, 4],
        embed_dim=1536,
        depth=24,
        num_heads=24,
        mlp_ratio=4,
    )
    encoder_kwargs.update(
        (k, kwargs[k]) for k in set(kwargs).intersection(encoder_kwargs))
    encoder_kwargs = {**encoder_kwargs, **kwargs}
    encoder = AudioTransformerMAE_Encoder(**encoder_kwargs)

    decoder_kwargs = dict(embed_dim=512,
                          depth=8,
                          num_heads=16,
                          input_dim=encoder_kwargs['embed_dim'],
                          outputdim=encoder.patch_embed.patch_size[0] *
                          encoder.patch_embed.patch_size[1],
                          num_patches=encoder.patch_embed.num_patches)
    decoder = AudioTransformerMAE_Decoder(**decoder_kwargs)
    return AudioTransformerMAE(encoder, decoder)


def dasheng_12B(**kwargs):
    encoder_kwargs = dict(
        patch_size=[64, 4],
        patch_stride=[64, 4],
        embed_dim=1536,
        depth=40,
        num_heads=24,
        mlp_ratio=4,
    )
    encoder_kwargs.update(
        (k, kwargs[k]) for k in set(kwargs).intersection(encoder_kwargs))
    encoder_kwargs = {**encoder_kwargs, **kwargs}
    encoder = AudioTransformerMAE_Encoder(**encoder_kwargs)

    decoder_kwargs = dict(embed_dim=768,
                          depth=8,
                          num_heads=24,
                          input_dim=encoder_kwargs['embed_dim'],
                          outputdim=encoder.patch_embed.patch_size[0] *
                          encoder.patch_embed.patch_size[1],
                          num_patches=encoder.patch_embed.num_patches)
    decoder = AudioTransformerMAE_Decoder(**decoder_kwargs)
    return AudioTransformerMAE(encoder, decoder)
